import os.path
from enum import Enum
from os import path
from typing import List

import typer
from incorta_cloud_client.api.storage_api import StorageApi
from incorta_cloud_client.model.forms_start_upload import FormsStartUpload
from incorta_cloud_client.model.forms_sync_storage import FormsSyncStorage
from incorta_cloud_client.model.forms_sync_storage_response_payload import FormsSyncStorageResponsePayload
from incorta_cloud_client.model.models_download_metadata import ModelsDownloadMetadata
from incorta_cloud_client.model.models_sync_storage_http_response import ModelsSyncStorageHttpResponse
from incorta_cloud_client.model.models_upload_metadata import ModelsUploadMetadata

from incorta_cloud.services.storage.upload_file import UploadFile
from typer import Typer

from incorta_cloud.services.storage.upload_tenant import UploadWorker
from .download_utils import DownloadWorker
from ...incorta_client import get_incorta_client, get_current_context

from ...utils import logger
from ...utils.adpter_utils import Adapter
from ...utils.concurrency.concurrency_manager import ConcurrencyManager
from ...utils.file_utils import FileUtils
from ...utils.string_utils import remove_leading_trailing_slash


class ConflictResolution(str, Enum):
    prefer_cloud = "prefer-cloud"
    prefer_local = "prefer-local"
    prefer_latest = "prefer-latest"


class SyncTenant:
    PREFER_CLOUD = 0
    PREFER_LOCAL = 1
    PREFER_LATEST = 2

    resolution_map = {
        ConflictResolution.prefer_cloud: PREFER_CLOUD,
        ConflictResolution.prefer_local: PREFER_LOCAL,
        ConflictResolution.prefer_latest: PREFER_LATEST,
    }

    def __init__(self, app: Typer) -> None:
        super().__init__()
        self.app = app

        @app.command()
        def sync_tenant_data(root_path: str = typer.Option(..., exists=True, file_okay=False, dir_okay=True),
                             conflict_resolution: ConflictResolution = typer.Option(..., ),
                             parallel_uploads: int = typer.Option(10, min=1, max=20),
                             parallel_downloads: int = typer.Option(10, min=1, max=20)) -> None:
            self.sync_tenant(root_path, conflict_resolution, parallel_uploads, parallel_downloads, storage_directory="data")

        @app.command()
        def sync_tenant_parquet(root_path: str = typer.Option(..., exists=True, file_okay=False, dir_okay=True, help="Tenant's folder root path"),
                                conflict_resolution: ConflictResolution = typer.Option(..., ),
                                parquet_dir: bool = typer.Option(False, help="Include Parquet directory in sync"),
                                compacted_dir: bool = typer.Option(False, help="Include Compacted directory in sync"),
                                parallel_uploads: int = typer.Option(10, min=1, max=20),
                                parallel_downloads: int = typer.Option(10, min=1, max=20)) -> None:
            if not parquet_dir and not compacted_dir:
                logger.error("One of parquet or compacted dirs must be enabled for the sync")
                return
            if parquet_dir:
                logger.info("Syncing Parquet Directory")
                self.sync_tenant(root_path, conflict_resolution, parallel_uploads, parallel_downloads, storage_directory="parquet")
            if compacted_dir:
                logger.info("Syncing Compacted Directory")
                self.sync_tenant(root_path, conflict_resolution, parallel_uploads, parallel_downloads, storage_directory="compacted")

    def sync_tenant(self, root_path: str, conflict_resolution: ConflictResolution, parallel_uploads: int, parallel_downloads: int, storage_directory="data"):
        file_generator = FileUtils.dir_files_generator(path.join(root_path, storage_directory))

        all_files = list(file_generator)
        local_files_data = [self.process_file(abs_path, "/"+storage_directory+rel_path, file_name) for (abs_path, rel_path, file_name) in all_files]

        sync_storage_response: FormsSyncStorageResponsePayload = self.sync_storage(local_files_data, conflict_resolution, storage_directory)
        root_path = os.path.abspath(root_path)
        self.upload_local_files(root_path, sync_storage_response.to_be_uploaded, parallel_uploads)
        self.download_local_files(root_path, sync_storage_response.to_be_downloaded, parallel_downloads)

    def sync_storage(self, local_files_data: List[FormsStartUpload], conflict_resolution: ConflictResolution, storage_directory="data") -> FormsSyncStorageResponsePayload:
        current_instance_id, current_instance_name, current_tenant_name = get_current_context()
        client = get_incorta_client()

        request_payload = FormsSyncStorage(
            local_objects=local_files_data,
            cluster_id=current_instance_id,
            tenant=current_tenant_name,
            conflict_resolution_type=SyncTenant.resolution_map[conflict_resolution],
            storage_directory=storage_directory
        )
        with client as api_client:
            api_instance = StorageApi(api_client)
            api_response: ModelsSyncStorageHttpResponse = api_instance.sync_storage(request_payload)
        logger.info("Found {} files to Upload, {} to Download".format(len(api_response.payload.to_be_uploaded), len(api_response.payload.to_be_downloaded)))

        return api_response.payload

    def process_file(self, abs_path, rel_path, file_name, ) -> FormsStartUpload:
        current_instance_id, current_instance_name, current_tenant_name = get_current_context()
        file_metadata = FileUtils.parse_file_metadata(file_name, abs_path, rel_path, current_instance_id, current_tenant_name)

        return Adapter.file_metadata_to_forms_start_upload(file_metadata)

    def upload_local_files(self, root_path, to_be_uploaded: List[ModelsUploadMetadata], parallel_uploads: int):
        uploader = UploadFile(self.app)
        client = get_incorta_client()

        thread_generator = (
            UploadWorker(uploader, client, os.path.join(root_path, remove_leading_trailing_slash(file.path), remove_leading_trailing_slash(file.name)), file.path, file.name,
                         file.upload_url) for file in to_be_uploaded)

        ConcurrencyManager.execute_in_parallel(thread_generator, parallel_uploads)

    def download_local_files(self, root_path, to_be_downloaded: List[ModelsDownloadMetadata], parallel_downloads: int):
        thread_generator = (DownloadWorker(root_path, file) for file in to_be_downloaded)

        ConcurrencyManager.execute_in_parallel(thread_generator, parallel_downloads)
