import os
import typing

import urllib3
import json

from incorta_cloud_client import ApiClient, ApiException
from incorta_cloud_client.api.upload_api import UploadApi
from incorta_cloud_client.model.models_upload_metadata import ModelsUploadMetadata
from rich.progress import TaskID

from incorta_cloud.incorta_client import get_current_context
from incorta_cloud.utils import logger
from incorta_cloud.utils.adpter_utils import Adapter
from incorta_cloud.utils.file_utils import BufferReader, FileUtils
from incorta_cloud.utils.logger.logger import Logger


def initiate_file_upload(name: str, file_aboslute_path: str, incorta_client: ApiClient, target_path: str = "") -> ModelsUploadMetadata:
    current_instance_id, current_instance_name, current_tenant_name = get_current_context()
    data = FileUtils.parse_file_metadata(
        name, file_aboslute_path, target_path, current_instance_id, current_tenant_name)
    forms_start_upload = Adapter.file_metadata_to_forms_start_upload(data)

    with incorta_client as api_client:
        api_instance = UploadApi(api_client)
        try:

            api_response = api_instance.initiate(forms_start_upload)
        except ApiException as e:
            if e.status == 401:
                logger.error("Missing or Expired Auth token, sign in first")
            elif e.status == 400:
                # logger.error(e)
                # logger.error("file already exists")
                return None
            else:
                logger.error(e)
            os._exit(1)

    return api_response.payload


def update_progress(task_id: TaskID, logger: Logger, already_uploaded: int = 0) -> typing.Callable[[int, int], typing.Any]:
    return lambda size, progress: logger.update_task_progress(Logger.upload_progress_name, task_id, progress + already_uploaded)


def upload_data(logger: Logger, data: str, upload_url: str, name: str = "") -> typing.Any:
    # pass
    http = urllib3.PoolManager()

    progresObj = logger.url_progress(
        Logger.upload_progress_name, "green")

    # first check the upload status to resume if possible
    r = http.request('PUT', upload_url,
                     headers={'Content-Type': 'application/json', 'Content-Length': 0,
                              'Content-Range': 'bytes */' + str(len(data))}
                     )

    if r.status == 200 or r.status == 201:
        # logger.info('\[{}] Upload was already finished.'.format(name))
        task_id = logger.add_complete_task_to_progress(
            Logger.upload_progress_name, upload_url, len(data), name, "Already Finished")

        return json.loads(r.data.decode('utf-8'))

    task_id = logger.add_task_to_progress(
        Logger.upload_progress_name, upload_url, len(data), name, "Uploading")

    start = 0
    headers = {'Content-Type': 'application/json', 'Content-Length': len(data)}
    if 'range' in r.headers:
        byte_range: str = r.headers['range']
        byte_range = byte_range.replace('bytes=', '')
        byte_range_list: typing.List[str] = byte_range.split('-')
        start = int(byte_range_list[1]) + 1
        end = len(data) - 1
        headers['Content-Length'] = str(end - start + 1)
        headers['Content-Range'] = "bytes {}-{}/{}".format(
            start, end, len(data))
    logger.start_task(Logger.upload_progress_name, task_id)
    logger.update_task_progress(Logger.upload_progress_name, task_id, start)
    body = BufferReader(data[start:], update_progress(task_id, logger, start))

    r = http.request('PUT', upload_url,
                     body=body,
                     headers=headers
                     )
    return json.loads(r.data.decode('utf-8'))
