import datetime
import hashlib
import io
import os
import datetime

from incorta_cloud.utils import logger
from incorta_cloud.utils.errors.internet_errors import CancelledError


class FileMetadata:
    def __init__(self, name, path, md5_hash, size, created_at, updated_at, cluster_id: str = None, tenant: str = None):
        self.tenant = tenant
        self.cluster_id = cluster_id
        self.md5_hash = md5_hash
        self.path = path
        self.name = name
        self.size = size
        self.created_at = created_at
        self.updated_at = updated_at


class FileUtils:
    @staticmethod
    def read_data(path):
        with open(path, 'rb') as f:
            data = f.read()
        return data

    @staticmethod
    def parse_file_metadata(name, absolute_path, target_path, current_instance_id, current_tenant_name) -> FileMetadata:
        file_data = FileUtils.read_data(absolute_path)
        md5_hash = hashlib.md5(file_data).hexdigest()
        # current_instance_id, current_instance_name, current_tenant_name = get_current_context()
        if current_instance_id is None:
            logger.error("Please set a context to run in first")
            os._exit(1)
        LOCAL_TIMEZONE = datetime.datetime.now(
            datetime.timezone.utc).astimezone().tzinfo
        return FileMetadata(
            name=name,
            md5_hash=md5_hash,
            path=target_path,
            size=len(file_data),
            cluster_id=current_instance_id,
            tenant=current_tenant_name,
            created_at=datetime.datetime.fromtimestamp(
                os.path.getctime(absolute_path), LOCAL_TIMEZONE).isoformat(),
            updated_at=datetime.datetime.fromtimestamp(
                os.path.getmtime(absolute_path), LOCAL_TIMEZONE).isoformat()
        )

    @staticmethod
    def dir_files_generator(root_dir_path):
        """

        Args:
            root_dir_path: the root directory to list all files in it

        Returns:
            tuple: (x,y.z) where x is the absolute path to the file, y is the relative path of the file to the dir_path, z is the file name
        """

        def find_relative_path(root_dir_path, dir_path, f_name):
            file_abs_path = os.path.abspath(os.path.join(dir_path, f_name))
            file_relative_path = file_abs_path.replace(os.path.abspath(
                root_dir_path), '')  # remove the slash at the begining
            file_relative_path_without_name = file_relative_path.replace(
                f_name, '')
            return file_relative_path_without_name

        return ((os.path.abspath(os.path.join(dir_path, f)), find_relative_path(root_dir_path, dir_path, f), f)
                for dir_path, dir_name, filenames in os.walk(root_dir_path)
                for f in filenames)

    @staticmethod
    def create_dirs(path):
        os.makedirs(path, exist_ok=True)


class BufferReader(io.BytesIO):
    def __init__(self, buf=b'',
                 callback=None,
                 cb_args=(),
                 cb_kwargs={}):
        self._callback = callback
        self._cb_args = cb_args
        self._cb_kwargs = cb_kwargs
        self._progress = 0
        self._len = len(buf)
        io.BytesIO.__init__(self, buf)

    def __len__(self):
        return self._len

    def read(self, n=-1):
        chunk = io.BytesIO.read(self, n)
        self._progress += int(len(chunk))
        self._cb_kwargs.update({
            'size': self._len,
            'progress': self._progress
        })
        if self._callback:
            try:
                self._callback(*self._cb_args, **self._cb_kwargs)
            except:  # catches exception from the callback
                raise CancelledError('The upload was cancelled.')
        return chunk
