import sys
import threading
import queue
import time
from typing import List, Any, Union, Dict

import typer
from rich.console import Console
from rich.live import Live
from rich.panel import Panel
from rich.progress import Progress, TextColumn, ProgressColumn, BarColumn, DownloadColumn, TransferSpeedColumn, TimeRemainingColumn, TaskID, SpinnerColumn
from rich.table import Table


class Logger(threading.Thread):
    console = Console()
    upload_progress_name = "Uploading"
    download_progress_name = "Downloading"

    def __init__(self) -> None:
        threading.Thread.__init__(self)
        super().__init__()
        self.console = Logger.console

        self.progressObjects: Dict[str, Progress] = dict()
        self.progress_table = Table.grid(expand=True)
        self.messageQueue = queue.Queue()
        self.live = Live(self.progress_table, refresh_per_second=10)
        # self.progressObj: Union[Progress, None] = None
        # self.start()

    def __enter__(self):
        self.start()

    def __exit__(self, exc_type, exc_val, exc_tb, sleep_time=3):
        time.sleep(sleep_time)
        if self.live.is_started:
            self.live.__exit__(None, None, None)
        self.messageQueue.put(lambda x: sys.exit(1))

    def run(self):
        while True:
            if len(self.progress_table.rows) > 0 and not self.live.is_started:
                self.live.__enter__()
            fn = self.messageQueue.get()
            fn(None)

    def log(self, message, style):
        if not self.progressObjects:
            self.console.log(message, style=style)
        else:
            self.messageQueue.put(
                lambda x: self.console.log(message, style=style))

    def debug(self, message):
        self.log(message, "cyan")

    def success(self, message):
        self.log(message, "green")

    def info(self, message):
        self.log(message, "")

    def error(self, message):
        self.log(message, "red")

    def critical(self, message):
        self.log(message, "bold red")

    def get_status_obj(self, message):
        return self.console.status(message)

    def progress(self, name: str, columns: List[ProgressColumn], ) -> Progress:
        if name not in self.progressObjects:
            self.progressObjects[name] = Progress(
                *columns
            )

        return self.progressObjects[name]

    def create_url_progress(self, name: str, color: str = None):
        if name in self.progressObjects:
            return
        self.progressObjects[name] = Progress(
            TextColumn("[bold blue]{task.fields[action]}", justify="right"),
            TextColumn("[bold blue]{task.fields[filename]}", justify="right"),
            BarColumn(bar_width=None),
            "[progress.percentage]{task.percentage:>3.1f}%",
            "•",
            DownloadColumn(),
            "•",
            TransferSpeedColumn(),
            "•",
            TimeRemainingColumn(),

            console=self.console
        )
        self.progress_table.add_row(
            Panel.fit(
                self.progressObjects[name], title=name, border_style=color if color is not None else "green", padding=(2, 2)
            ),
        )

    def url_progress(self, name: str, color: str = None) -> Progress:
        self.create_url_progress(name, color)
        return self.progressObjects[name]

    def add_complete_task_to_progress(self, progress_name: str, url: str, total_length: int, filename: str, action: str) -> Any:
        task_id = self.add_task_to_progress(
            progress_name, url, total_length, filename, action)
        self.update_task_progress(progress_name, task_id, total_length)
        return task_id

    def add_task_to_progress(self, progress_name: str, url: str, total_length: int, filename: str, action: str) -> Any:
        if progress_name not in self.progressObjects:
            return None
        progress_object = self.progressObjects[progress_name]

        # This will break if the response doesn't contain content length
        task_id = progress_object.add_task(
            action, action=action, filename=filename, start=True, total=total_length)
        # self.messageQueue.put(lambda x: None) # refresh view if live is not loaded yet
        return task_id

    def start_task(self, task_id, progress_name: str):
        if progress_name not in self.progressObjects:
            return
        self.progressObjects[progress_name].start_task(task_id=task_id)

    def stop_task(self, task_id):
        pass
        # self.progressObj.stop_task(task_id=task_id)
        # self.progressObj.remove_task(task_id=task_id)

    def update_task_progress(self, progress_name, task_id: TaskID, completed_length: int) -> None:
        if progress_name not in self.progressObjects:
            return
        self.messageQueue.put(lambda x: self.progressObjects[progress_name].update(
            task_id, completed=completed_length))
